import asyncio
import logging
from aioscrapy.db import db_manager

logger = logging.getLogger(__name__)


class MongoPipeline:
    def __init__(self, settings):
        self.cache_num = settings.getint('SAVE_CACHE_NUM', 500)
        self.save_cache_interval = settings.getint('SAVE_CACHE_INTERVAL', 10)
        self.lock = asyncio.Lock()
        self.running: bool = True
        self.db_alias_cache = {}
        self.table_cache = {}
        self.item_cache = {}
        self.db_cache = {}

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    async def open_spider(self, spider):
        asyncio.create_task(self.save_heartbeat())

    async def save_heartbeat(self):
        while self.running:
            await asyncio.sleep(self.save_cache_interval)
            asyncio.create_task(self.save_all())

    async def process_item(self, item, spider):
        await self.save_item(item)
        return item

    async def close_spider(self, spider):
        self.running = False
        await self.save_all()

    def parse_item_to_cache(self, item: dict):
        item.pop('save_insert_type', None)
        db_name = item.pop('save_db_name', None)
        table_name = item.pop('save_table_name', None)
        assert table_name is not None, Exception('please set save_table_name')
        save_db_alias = item.pop('save_db_alias', ['default'])
        if isinstance(save_db_alias, str):
            save_db_alias = [save_db_alias]

        cache_key = ''.join(save_db_alias) + (db_name or '') + table_name

        if self.table_cache.get(cache_key) is None:
            self.db_alias_cache[cache_key] = save_db_alias
            self.table_cache[cache_key] = table_name
            self.db_cache[cache_key] = db_name
            self.item_cache[cache_key] = []

        self.item_cache[cache_key].append(item)
        return cache_key, len(self.item_cache[cache_key])

    async def save_all(self):
        async with self.lock:
            for cache_key, items in self.item_cache.items():
                items and await self._save(cache_key)

    async def save_item(self, item: dict):
        async with self.lock:
            cache_key, cache_count = self.parse_item_to_cache(item)
            if cache_count >= self.cache_num:
                await self._save(cache_key)

    async def _save(self, cache_key):
        table_name = self.table_cache[cache_key]
        try:
            for alias in self.db_alias_cache[cache_key]:
                try:
                    executor = db_manager.mongo.executor(alias)
                    result = await executor.insert(
                        table_name, self.item_cache[cache_key], db_name=self.db_cache[cache_key]
                    )
                    logger.info(
                        f'table:{alias}->{table_name} sum:{len(self.item_cache[cache_key])} ok:{len(result.inserted_ids)}')
                except Exception as e:
                    logger.exception(f'save data error, table:{alias}->{table_name}, err_msg:{e}')
        finally:
            self.item_cache[cache_key] = []




import os
import shutil
import sys
import tempfile
from contextlib import contextmanager

import pkg_resources

try:
    from scrapyd import get_application
    from scrapyd.interfaces import IEggStorage
except ImportError:
    pass


def activate_egg(eggpath):
    """Activate a Scrapy egg file. This is meant to be used from egg runners
    to activate a Scrapy egg file. Don't use it from other code as it may
    leave unwanted side effects.
    """
    try:
        d = next(pkg_resources.find_distributions(eggpath))
    except StopIteration:
        raise ValueError("Unknown or corrupt egg")
    d.activate()
    settings_module = d.get_entry_info('aioscrapy', 'settings').module_name
    os.environ.setdefault('AIOSCRAPY_SETTINGS_MODULE', settings_module)


@contextmanager
def project_environment(project):
    app = get_application()
    eggstorage = app.getComponent(IEggStorage)
    eggversion = os.environ.get('AIOSCRAPY_EGG_VERSION', None)
    version, eggfile = eggstorage.get(project, eggversion)
    if eggfile:
        prefix = '%s-%s-' % (project, version)
        fd, eggpath = tempfile.mkstemp(prefix=prefix, suffix='.egg')
        lf = os.fdopen(fd, 'wb')
        shutil.copyfileobj(eggfile, lf)
        lf.close()
        activate_egg(eggpath)
    else:
        eggpath = None
    try:
        assert 'aioscrapy.conf' not in sys.modules, "aioscrapy settings already loaded"
        yield
    finally:
        if eggpath:
            os.remove(eggpath)


def main():
    os.environ.update({f'AIO{k}': v for k, v in os.environ.items() if k.startswith('SCRAPY_')})

    project = os.environ['AIOSCRAPY_PROJECT']
    with project_environment(project):
        from aioscrapy.cmdline import execute
        execute()


if __name__ == '__main__':
    main()



from abc import ABCMeta, abstractmethod

import aioscrapy


class AbsDBPoolManager(object, metaclass=ABCMeta):

    @property
    @abstractmethod
    def _clients(self) -> dict:
        return {}

    @abstractmethod
    async def create(self, alias: str, params: dict):
        """Create pool of connection"""

    @abstractmethod
    def get_pool(self, alias: str):
        """Get pool named `alias`"""

    @abstractmethod
    async def close_all(self):
        """Close all pool"""

    @abstractmethod
    async def close(self, alias: str):
        """Close pool named `alias`"""

    @abstractmethod
    async def from_dict(self, db_args: dict):
        """Create pool with dict"""

    @abstractmethod
    async def from_settings(self, settings: aioscrapy.Settings):
        """Create pool with settings"""

    async def from_crawler(self, crawler: "aioscrapy.Crawler"):
        return await self.from_settings(crawler.settings)

    def __call__(self, alias: str):
        return self.get_pool(alias)

    def __getitem__(self, alias: str):
        return self.get_pool(alias)

    def __getattr__(self, alias: str):
        return self.get_pool(alias)



from abc import ABCMeta, abstractmethod

from aioscrapy import Request, Spider


class DupeFilterBase(metaclass=ABCMeta):
    """Request Fingerprint duplicates filter"""

    @classmethod
    @abstractmethod
    def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        """ Get Instance of RFPDupeFilter from crawler """

    @abstractmethod
    async def exist_fingerprint(self, request: Request) -> bool:
        """ Check whether fingerprint of request exists """

    @abstractmethod
    async def add_fingerprint(self, request: Request) -> None:
        """ Add fingerprint of request """

    @abstractmethod
    def close(self, reason: str = '') -> None:
        """ Delete data on close """

    @abstractmethod
    def log(self, request: Request, spider: Spider) -> None:
        """ Logs given request """



from redis.asyncio import BlockingConnectionPool, Redis

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager


class RedisExecutor:

    def __init__(self, alias: str, pool_manager: "AioRedisPoolManager"):
        self.alias = alias
        self.pool_manager = pool_manager

    def __getattr__(self, command: str):
        redis_pool: Redis = self.pool_manager.get_pool(self.alias)
        return getattr(redis_pool, command)


class AioRedisPoolManager(AbsDBPoolManager):
    _clients = {}

    async def create(self, alias: str, params: dict) -> Redis:
        """Create Redis client"""
        if alias in self._clients:
            return self._clients[alias]

        params = params.copy()
        url = params.pop('url', None)
        params.setdefault('socket_connect_timeout', 30)
        if url:
            connection_pool = BlockingConnectionPool.from_url(url, **params)
        else:
            connection_pool = BlockingConnectionPool(**params)
        redis = Redis(connection_pool=connection_pool)
        return self._clients.setdefault(alias, redis)

    def get_pool(self, alias: str):
        """Get redis client"""
        redis_pool: Redis = self._clients.get(alias)
        assert redis_pool is not None, f"Dont create the redis client named {alias}"
        return redis_pool

    def executor(self, alias: str) -> RedisExecutor:
        """Get RedisExecutor"""
        return RedisExecutor(alias, self)

    async def close(self, alias: str):
        """Close redis pool named `alias`"""
        redis = self._clients.pop(alias, None)
        if redis:
            await redis.close()
            await redis.connection_pool.disconnect()

    async def close_all(self):
        """Close all clients of redis"""
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        """Create redis with dict"""
        for alias, redis_args in db_args.items():
            await self.create(alias, redis_args)

    async def from_settings(self, settings: aioscrapy.Settings):
        """Create redis with settings"""
        for alias, redis_args in settings.getdict('REDIS_ARGS').items():
            await self.create(alias, redis_args)


redis_manager = AioRedisPoolManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        await redis_manager.create('default', {
            'url': 'redis://@192.168.234.128:6379/9',
        })
        redis = redis_manager.executor('default')
        print(await redis.zadd('key1', {'value': 2}))

        async with redis.pipeline(transaction=True) as pipe:
            results, count = await (
                pipe.zrange('key1', 0, 0)
                    .zremrangebyrank('key1', 0, 0)
                    .execute()
            )

        print(results)
        await redis_manager.close_all()


    # asyncio.run(test())
    asyncio.get_event_loop().run_until_complete(test())



"""Helper functions for working with signals"""
import asyncio
import logging

from pydispatch.dispatcher import Anonymous, Any, disconnect, getAllReceivers, liveReceivers
from pydispatch.robustapply import robustApply
from aioscrapy.exceptions import StopDownload

logger = logging.getLogger(__name__)


class _IgnoredException(Exception):
    pass


async def robustApplyWrap(f, recv, *args, **kw):
    dont_log = kw.pop('dont_log', None)
    spider = kw.get('spider', None)
    try:
        result = f(recv, *args, **kw)
        if asyncio.iscoroutine(result):
            return await result
    except (Exception, BaseException) as exc:  # noqa: E722
        if dont_log is None or not isinstance(exc, dont_log):
            logger.error("Error caught on signal handler: %(receiver)s",
                         {'receiver': recv},
                         exc_info=exc,
                         extra={'spider': spider})
        return exc


async def send_catch_log(signal=Any, sender=Anonymous, *arguments, **named):
    """Like pydispatcher.robust.sendRobust but it also logs errors and returns
    Failures instead of exceptions.
    """
    named['dont_log'] = (named.pop('dont_log', _IgnoredException), StopDownload)
    responses = []
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        result = await robustApplyWrap(robustApply, receiver, signal=signal, sender=sender, *arguments, **named)
        responses.append((receiver, result))
    return responses


async def send_catch_log_deferred(signal=Any, sender=Anonymous, *arguments, **named):
    """Like send_catch_log but supports returning deferreds on signal handlers.
    Returns a deferred that gets fired once all signal handlers deferreds were
    fired.
    """
    dfds = []
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        dfds.append(asyncio.create_task(
            robustApplyWrap(robustApply, receiver, signal=signal, sender=sender, *arguments, **named)))
    res = await asyncio.gather(*dfds)
    return res


def disconnect_all(signal=Any, sender=Any):
    """Disconnect all signal handlers. Useful for cleaning up after running
    tests
    """
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        disconnect(receiver, signal=signal, sender=sender)




"""
This module implements the Request class which is used to represent HTTP
requests in Aioscrapy.

See documentation in docs/topics/request-response.rst
"""
import hashlib
import inspect
import json
from typing import Callable, List, Optional, Tuple, Type, TypeVar

from w3lib.url import canonicalize_url
from w3lib.url import safe_url_string

import aioscrapy
from aioscrapy.http.headers import Headers
from aioscrapy.utils.curl import curl_to_request_kwargs
from aioscrapy.utils.python import to_unicode
from aioscrapy.utils.url import escape_ajax

RequestTypeVar = TypeVar("RequestTypeVar", bound="Request")


class Request(object):
    attributes: Tuple[str, ...] = (
        "url", "callback", "method", "headers", "body",
        "cookies", "meta", "encoding", "priority",
        "dont_filter", "errback", "flags", "cb_kwargs",
        "fingerprint"
    )

    def __init__(
            self,
            url: str,
            callback: Optional[Callable] = None,
            method: str = 'GET',
            headers: Optional[dict] = None,
            body: Optional[str] = None,
            cookies: Optional[dict] = None,
            meta: Optional[dict] = None,
            encoding: str = 'utf-8',
            priority: int = 0,
            dont_filter: bool = False,
            errback: Optional[Callable] = None,
            flags: Optional[List[str]] = None,
            cb_kwargs: Optional[Callable] = None,
            fingerprint: Optional[str] = None,
    ):

        self._encoding = encoding
        self.method = str(method).upper()
        self._set_url(url)
        self._set_body(body)
        assert isinstance(priority, int), f"Request priority not an integer: {priority!r}"
        self.priority = priority

        self.callback = callback
        self.errback = errback

        self.cookies = cookies or {}
        self.headers = Headers(headers or {})
        self.dont_filter = dont_filter
        self._fingerprint = fingerprint

        self._meta = dict(meta) if meta else None
        self._cb_kwargs = dict(cb_kwargs) if cb_kwargs else None
        self.flags = [] if flags is None else list(flags)

    @property
    def cb_kwargs(self) -> dict:
        if self._cb_kwargs is None:
            self._cb_kwargs = {}
        return self._cb_kwargs

    @property
    def meta(self) -> dict:
        if self._meta is None:
            self._meta = {}
        return self._meta

    def _get_url(self) -> str:
        return self._url

    def _set_url(self, url: str) -> None:
        if not isinstance(url, str):
            raise TypeError(f'Request url must be str or unicode, got {type(url).__name__}')

        s = safe_url_string(url, self.encoding)
        self._url = escape_ajax(s)

        if (
                '://' not in self._url
                and not self._url.startswith('about:')
                and not self._url.startswith('data:')
        ):
            raise ValueError(f'Missing scheme in request url: {self._url}')

    url = property(_get_url, _set_url)

    def _get_body(self) -> str:
        return self._body

    def _set_body(self, body: str) -> None:
        self._body = '' if body is None else body

    body = property(_get_body, _set_body)

    def _set_fingerprint(self, fingerprint: str) -> None:
        self._fingerprint = fingerprint

    def _get_fingerprint(self) -> str:
        if self._fingerprint is None:
            self._fingerprint = self.make_fingerprint()
        return self._fingerprint

    fingerprint = property(_get_fingerprint, _set_fingerprint)

    @property
    def encoding(self) -> str:
        return self._encoding

    def __str__(self) -> str:
        return f"<{self.method} {self.url}>"

    __repr__ = __str__

    def copy(self) -> "Request":
        """Return a copy of this Request"""
        return self.replace()

    def replace(self, *args, **kwargs) -> "Request":
        """Create a new Request with the same attributes except for those given new values."""
        for x in self.attributes:
            kwargs.setdefault(x, getattr(self, x))
        cls = kwargs.pop('cls', self.__class__)
        return cls(*args, **kwargs)

    @classmethod
    def from_curl(
            cls: Type[RequestTypeVar], curl_command: str, ignore_unknown_options: bool = True, **kwargs
    ) -> RequestTypeVar:
        """Create a Request object from a string containing a `cURL"""
        request_kwargs = curl_to_request_kwargs(curl_command, ignore_unknown_options)
        request_kwargs.update(kwargs)
        return cls(**request_kwargs)

    def make_fingerprint(
            self,
            keep_fragments: bool = False,
    ) -> str:
        """ make the request fingerprint. """
        return hashlib.sha1(
            json.dumps({
                'method': to_unicode(self.method),
                'url': canonicalize_url(self.url, keep_fragments=keep_fragments),
                'body': self.body,
            }, sort_keys=True).encode()
        ).hexdigest()

    def to_dict(self, *, spider: Optional["aioscrapy.Spider"] = None) -> dict:
        """Return a dictionary containing the Request's data.

        Use :func:`~scrapy.utils.request.request_from_dict` to convert back into a :class:`~scrapy.Request` object.

        If a spider is given, this method will try to find out the name of the spider methods used as callback
        and errback and include them in the output dict, raising an exception if they cannot be found.
        """
        d = {
            "url": self.url,  # urls are safe (safe_string_url)
            "callback": _find_method(spider, self.callback) if callable(self.callback) else self.callback,
            "errback": _find_method(spider, self.errback) if callable(self.errback) else self.errback,
            "headers": dict(self.headers),
        }
        if self._fingerprint:
            d['fingerprint'] = self._fingerprint

        for attr in self.attributes:
            if attr != 'fingerprint':
                d.setdefault(attr, getattr(self, attr))
        if type(self) is not Request:
            d["_class"] = self.__module__ + '.' + self.__class__.__name__
        return d


def _find_method(obj, func):
    """Helper function for Request.to_dict"""
    # Only instance methods contain ``__func__``
    if obj and hasattr(func, '__func__'):
        members = inspect.getmembers(obj, predicate=inspect.ismethod)
        for name, obj_func in members:
            # We need to use __func__ to access the original function object because instance
            # method objects are generated each time attribute is retrieved from instance.
            #
            # Reference: The standard type hierarchy
            # https://docs.python.org/3/reference/datamodel.html
            if obj_func.__func__ is func.__func__:
                return name
    raise ValueError(f"Function {func} is not an instance method in: {obj}")



"""
aioscrapy extension for collecting scraping stats
"""
import pprint
import logging

logger = logging.getLogger(__name__)


class StatsCollector:

    def __init__(self, crawler):
        self._dump = crawler.settings.getbool('STATS_DUMP')
        self._stats = {}

    def get_value(self, key, default=None, spider=None):
        return self._stats.get(key, default)

    def get_stats(self, spider=None):
        return self._stats

    def set_value(self, key, value, spider=None):
        self._stats[key] = value

    def set_stats(self, stats, spider=None):
        self._stats = stats

    def inc_value(self, key, count=1, start=0, spider=None):
        d = self._stats
        d[key] = d.setdefault(key, start) + count

    def max_value(self, key, value, spider=None):
        self._stats[key] = max(self._stats.setdefault(key, value), value)

    def min_value(self, key, value, spider=None):
        self._stats[key] = min(self._stats.setdefault(key, value), value)

    def clear_stats(self, spider=None):
        self._stats.clear()

    def open_spider(self, spider):
        pass

    def close_spider(self, spider, reason):
        if self._dump:
            logger.info("Dumping aioscrapy stats:\n" + pprint.pformat(self._stats),
                        extra={'spider': spider})
        self._persist_stats(self._stats, spider)

    def _persist_stats(self, stats, spider):
        pass


class MemoryStatsCollector(StatsCollector):

    def __init__(self, crawler):
        super().__init__(crawler)
        self.spider_stats = {}

    def _persist_stats(self, stats, spider):
        self.spider_stats[spider.name] = stats


class DummyStatsCollector(StatsCollector):

    def get_value(self, key, default=None, spider=None):
        return default

    def set_value(self, key, value, spider=None):
        pass

    def set_stats(self, stats, spider=None):
        pass

    def inc_value(self, key, count=1, start=0, spider=None):
        pass

    def max_value(self, key, value, spider=None):
        pass

    def min_value(self, key, value, spider=None):
        pass



from abc import ABCMeta, abstractmethod
from typing import Optional, Any, List

import aioscrapy
from aioscrapy.serializer import AbsSerializer
from aioscrapy.utils.reqser import request_from_dict


class AbsQueue(metaclass=ABCMeta):
    """Per-spider base queue class"""

    def __init__(
            self,
            container: Any,
            spider: Optional[aioscrapy.Spider] = None,
            key: Optional[str] = None,
            serializer: Optional[AbsSerializer] = None
    ) -> None:
        """Initialize per-spider queue"""
        self.container = container
        self.spider = spider
        self.key = key
        self.serializer = serializer

    @property
    @abstractmethod
    def inc_key(self) -> str:
        """stats inc_value"""

    @classmethod
    @abstractmethod
    async def from_spider(cls, spider: aioscrapy.Spider) -> "AbsQueue":
        """get queue instance from spider"""

    def _encode_request(self, request: aioscrapy.Request) -> Any:
        """Encode a request object"""
        obj = request.to_dict(spider=self.spider)
        return self.serializer.dumps(obj)

    def _decode_request(self, encoded_request: Any) -> aioscrapy.Request:
        """Decode an request previously encoded"""
        obj = self.serializer.loads(encoded_request)
        return request_from_dict(obj, spider=self.spider)

    def __len__(self) -> None:
        """Return the length of the queue"""
        raise Exception('please use len()')

    @abstractmethod
    async def len(self) -> int:
        """Return the length of the queue"""

    @abstractmethod
    async def push(self, request: aioscrapy.Request) -> None:
        """Push a request"""

    @abstractmethod
    async def push_batch(self, requests: List[aioscrapy.Request]) -> None:
        """Push a batch requests"""

    @abstractmethod
    async def pop(self, timeout: int = 0) -> Optional[aioscrapy.Request]:
        """Pop a request"""

    @abstractmethod
    async def clear(self) -> None:
        """Clear queue/stack"""



from abc import abstractmethod
from typing import Optional, Type, TypeVar, List

import aioscrapy
from aioscrapy.queue import AbsQueue
from aioscrapy.statscollectors import StatsCollector
from aioscrapy.utils.misc import load_instance
from aioscrapy.utils.tools import call_helper


class BaseSchedulerMeta(type):
    """
    Metaclass to check scheduler classes against the necessary interface
    """

    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (
                hasattr(subclass, "has_pending_requests") and callable(subclass.has_pending_requests)
                and hasattr(subclass, "enqueue_request") and callable(subclass.enqueue_request)
                and hasattr(subclass, "enqueue_request_batch") and callable(subclass.enqueue_request_batch)
                and hasattr(subclass, "next_request") and callable(subclass.next_request)
        )


class BaseScheduler(metaclass=BaseSchedulerMeta):

    @classmethod
    async def from_crawler(cls, crawler: "aioscrapy.Crawler") -> "BaseScheduler":
        """
        Factory method which receives the current :class:`~scrapy.crawler.Crawler` object as argument.
        """
        return cls()

    async def close(self, reason: str) -> None:
        """
        Called when the spider is closed by the engine. It receives the reason why the crawl
        finished as argument and it's useful to execute cleaning code.

        :param reason: a string which describes the reason why the spider was closed
        :type reason: :class:`str`
        """
        pass

    @abstractmethod
    async def has_pending_requests(self) -> bool:
        """
        ``True`` if the scheduler has enqueued requests, ``False`` otherwise
        """
        raise NotImplementedError()

    @abstractmethod
    async def enqueue_request_batch(self, requests: List[aioscrapy.Request]) -> bool:
        """
        Process a batch requests received by the engine.

        Return ``True`` if the request is stored correctly, ``False`` otherwise.

        If ``False``, the engine will fire a ``request_dropped`` signal, and
        will not make further attempts to schedule the request at a later time.
        For reference, the default Scrapy scheduler returns ``False`` when the
        request is rejected by the dupefilter.
        """
        raise NotImplementedError()

    @abstractmethod
    async def enqueue_request(self, request: aioscrapy.Request) -> bool:
        """
        Process a request received by the engine.

        Return ``True`` if the request is stored correctly, ``False`` otherwise.

        If ``False``, the engine will fire a ``request_dropped`` signal, and
        will not make further attempts to schedule the request at a later time.
        For reference, the default Scrapy scheduler returns ``False`` when the
        request is rejected by the dupefilter.
        """
        raise NotImplementedError()

    @abstractmethod
    async def next_request(self) -> Optional[aioscrapy.Request]:
        """
        Return the next :class:`~scrapy.http.Request` to be processed, or ``None``
        to indicate that there are no requests to be considered ready at the moment.

        Returning ``None`` implies that no request from the scheduler will be sent
        to the downloader in the current reactor cycle. The engine will continue
        calling ``next_request`` until ``has_pending_requests`` is ``False``.
        """
        raise NotImplementedError()


SchedulerTV = TypeVar("SchedulerTV", bound="Scheduler")


class Scheduler(BaseScheduler):

    def __init__(
            self,
            queue: AbsQueue,
            spider: aioscrapy.Spider,
            stats=Optional[StatsCollector],
            persist: bool = True
    ):
        self.queue = queue
        self.spider = spider
        self.stats = stats
        self.persist = persist

    @classmethod
    async def from_crawler(cls: Type[SchedulerTV], crawler: "aioscrapy.Crawler") -> SchedulerTV:
        instance = cls(
            await load_instance(crawler.settings['SCHEDULER_QUEUE_CLASS'], spider=crawler.spider),
            crawler.spider,
            stats=crawler.stats,
            persist=crawler.settings.getbool('SCHEDULER_PERSIST', True)
        )

        if crawler.settings.getbool('SCHEDULER_FLUSH_ON_START', False):
            await instance.flush()

        count = await call_helper(instance.queue.len)
        count and crawler.spider.log("Resuming crawl (%d requests scheduled)" % count)

        return instance

    async def close(self, reason: str) -> None:
        if not self.persist:
            await self.flush()

    async def flush(self) -> None:
        await call_helper(self.queue.clear)

    async def enqueue_request_batch(self, requests: List[aioscrapy.Request]) -> bool:
        await call_helper(self.queue.push_batch, requests)
        if self.stats:
            self.stats.inc_value(self.queue.inc_key, count=len(requests), spider=self.spider)
        return True

    async def enqueue_request(self, request: aioscrapy.Request) -> bool:
        await call_helper(self.queue.push, request)
        if self.stats:
            self.stats.inc_value(self.queue.inc_key, spider=self.spider)
        return True

    async def next_request(self, count: int = 1) -> Optional[aioscrapy.Request]:
        async for request in self.queue.pop(count):
            if request and self.stats:
                self.stats.inc_value(self.queue.inc_key, spider=self.spider)
            yield request

    async def has_pending_requests(self) -> bool:
        return await call_helper(self.queue.len) > 0



import logging
import socket
from contextlib import asynccontextmanager

from aiomysql import create_pool

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager

logger = logging.getLogger(__name__)


class MysqlExecutor:
    def __init__(self, alias: str, pool_manager: "AioMysqlPoolManager"):
        self.alias = alias
        self.pool_manager = pool_manager

    async def insert(self, sql, value):
        async with self.pool_manager.get(self.alias) as (connect, cursor):
            try:
                result = await cursor.executemany(sql, value)
                await connect.commit()
                return result
            except Exception as e:
                await connect.rollback()
                raise Exception from e

    async def execute(self, sql: str):
        async with self.pool_manager.get(self.alias) as (connect, cursor):
            await cursor.execute(sql)
            return await cursor.fetchall()

    async def query(self, sql: str):
        return await self.execute(sql)


class AioMysqlPoolManager(AbsDBPoolManager):
    _clients = {}

    async def create(self, alias: str, params: dict):
        if alias in self._clients:
            return self._clients[alias]

        params = params.copy()
        # When the host is domain,converted to IP
        # https://github.com/aio-libs/aiomysql/issues/641
        params.update({'host': socket.gethostbyname(params['host'])})
        params.setdefault('connect_timeout', 30)
        mysql_pool = await create_pool(**params)
        return self._clients.setdefault(alias, mysql_pool)

    def get_pool(self, alias: str):
        mysql_pool = self._clients.get(alias)
        assert mysql_pool is not None, f"Dont create the mysql pool named {alias}"
        return mysql_pool

    @asynccontextmanager
    async def get(self, alias: str, ping=False):
        """Get connection and cursor of mysql"""
        mysql_pool = self.get_pool(alias)
        conn = await mysql_pool.acquire()
        cur = await conn.cursor()
        try:
            if ping:
                await conn.ping()
            yield conn, cur
        finally:
            await cur.close()
            await mysql_pool.release(conn)

    def executor(self, alias: str) -> MysqlExecutor:
        return MysqlExecutor(alias, self)

    async def close(self, alias: str):
        mysql_pool = self._clients.pop(alias, None)
        if mysql_pool:
            mysql_pool.close()
            await mysql_pool.wait_closed()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        for alias, mysql_args in db_args.items():
            await self.create(alias, mysql_args)

    async def from_settings(self, settings: aioscrapy.Settings):
        for alias, mysql_args in settings.getdict('MYSQL_ARGS').items():
            await self.create(alias, mysql_args)


mysql_manager = AioMysqlPoolManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        mysql_pool = await mysql_manager.create('default', {
            'db': 'mysql',
            'user': 'root',
            'password': '123456',
            'host': '192.168.234.128',
            'port': 3306
        })

        # 方式一:
        try:
            conn = await mysql_pool.acquire()
            cur = await conn.cursor()
            await cur.execute('select * from user')
            print(await cur.fetchall())
            # await conn.commit()
        finally:
            await cur.close()
            await mysql_pool.release(conn)

        # 方式二:
        async with mysql_manager.get('xxx') as (conn, cur):
            print(await cur.execute('select 1'))
            # await conn.commit()

        await mysql_manager.close_all()


    asyncio.run(test())



from asyncio import PriorityQueue, Queue, LifoQueue
from asyncio.queues import QueueEmpty
from typing import Optional

import aioscrapy
from aioscrapy.queue import AbsQueue
from aioscrapy.serializer import AbsSerializer
from aioscrapy.utils.misc import load_object


class MemoryQueueBase(AbsQueue):
    inc_key = 'scheduler/enqueued/memory'

    def __init__(
            self,
            container: Queue,
            spider: Optional[aioscrapy.Spider],
            key: Optional[str] = None,
            serializer: Optional[AbsSerializer] = None,
            max_size: int = 0
    ) -> None:
        super().__init__(container, spider, key, serializer)
        self.max_size = max_size

    @classmethod
    async def from_spider(cls, spider: aioscrapy.Spider) -> "MemoryQueueBase":
        max_size: int = spider.settings.getint("QUEUE_MAXSIZE", 0)
        queue: Queue = cls.get_queue(max_size)
        queue_key: str = spider.settings.get("SCHEDULER_QUEUE_KEY", '%(spider)s:requests')
        serializer: str = spider.settings.get("SCHEDULER_SERIALIZER", "aioscrapy.serializer.PickleSerializer")
        serializer: AbsSerializer = load_object(serializer)
        return cls(
            queue,
            spider,
            queue_key % {'spider': spider.name},
            serializer=serializer
        )

    def len(self) -> int:
        """Return the length of the queue"""
        return self.container.qsize()

    @staticmethod
    def get_queue(max_size: int) -> Queue:
        raise NotImplementedError

    async def push(self, request) -> None:
        data = self._encode_request(request)
        await self.container.put(data)

    async def push_batch(self, requests) -> None:
        for request in requests:
            await self.push(request)

    async def pop(self, count: int = 1) -> None:
        for _ in range(count):
            try:
                data = self.container.get_nowait()
            except QueueEmpty:
                break
            yield self._decode_request(data)

    async def clear(self, timeout: int = 0) -> None:
        self.container = self.get_queue(self.max_size)


class MemoryFifoQueue(MemoryQueueBase):

    @staticmethod
    def get_queue(max_size: int) -> Queue:
        return Queue(max_size)


class MemoryLifoQueue(MemoryFifoQueue):
    @staticmethod
    def get_queue(max_size: int) -> LifoQueue:
        return LifoQueue(max_size)


class MemoryPriorityQueue(MemoryFifoQueue):
    @staticmethod
    def get_queue(max_size: int) -> PriorityQueue:
        return PriorityQueue(max_size)

    async def push(self, request: aioscrapy.Request) -> None:
        data = self._encode_request(request)
        score = request.priority
        await self.container.put((score, data))

    async def pop(self, count: int = 1) -> Optional[aioscrapy.Request]:
        for _ in range(count):
            try:
                score, data = self.container.get_nowait()
            except QueueEmpty:
                break
            yield self._decode_request(data)


SpiderQueue = MemoryFifoQueue
SpiderStack = MemoryLifoQueue
SpiderPriorityQueue = MemoryPriorityQueue



import logging
import os
from typing import Optional, Set

from aioscrapy import Request, Spider
from aioscrapy.dupefilters import DupeFilterBase
from aioscrapy.utils.request import referer_str


class DiskRFPDupeFilter(DupeFilterBase):
    """Request Fingerprint duplicates filter built with Disk storage"""

    def __init__(self, path: Optional[str] = None, debug: bool = False):
        self.file: Optional["File object"] = None
        self.debug = debug
        self.fingerprints: Set = set()
        self.logdupes: bool = True
        self.logger = logging.getLogger(__name__)
        if path:
            self.file = open(os.path.join(path, 'requests.seen'), 'a+')
            self.file.seek(0)
            self.fingerprints.update(x.rstrip() for x in self.file)

    @classmethod
    def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG')
        path = crawler.settings.get('JOBDIR', './job_dir')
        if path and not os.path.exists(path):
            os.makedirs(path)
        return cls(path, debug)

    async def exist_fingerprint(self, request: Request) -> bool:
        return request.fingerprint in self.fingerprints

    async def add_fingerprint(self, request: Request) -> None:
        self.fingerprints.add(request.fingerprint)
        if self.file:
            self.file.write(request.fingerprint + '\n')

    def close(self, reason: str = '') -> None:
        if self.file:
            self.file.close()

    def log(self, request: Request, spider: Spider):
        if self.debug:
            msg = "Filtered duplicate request: %(request)s (referer: %(referer)s)"
            args = {'request': request, 'referer': referer_str(request)}
            self.logger.debug(msg, args, extra={'spider': spider})
        elif self.logdupes:
            msg = ("Filtered duplicate request: %(request)s"
                   " - no more duplicates will be shown"
                   " (see DUPEFILTER_DEBUG to show all duplicates)")
            self.logger.debug(msg, {'request': request}, extra={'spider': spider})
            self.logdupes = False

        spider.crawler.stats.inc_value('dupefilter/filtered', spider=spider)


RFPDupeFilter = DiskRFPDupeFilter



"""Helper functions which don't fit anywhere else"""

from importlib import import_module
from pkgutil import iter_modules

from aioscrapy.utils.tools import call_helper


def walk_modules(path):
    """Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.

    For example: walk_modules('aioscrapy.utils')
    """

    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def load_object(path: str):
    """Load an object given its absolute object path, and return it.

    The object can be the import path of a class, function, variable or an
    instance, e.g. 'aioscrapy.libs.downloader.redirect.RedirectMiddleware'..
    """
    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError(f"Error loading object '{path}': not a full path")

    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError(f"Module '{module}' doesn't define any object named '{name}'")

    return obj


async def create_instance(objcls, settings, crawler, *args, spider=None, **kwargs):
    """Construct a class instance using its ``from_crawler`` or
    ``from_settings`` constructors, if available.

    At least one of ``settings`` and ``crawler`` needs to be different from
    ``None``. If ``settings `` is ``None``, ``crawler.settings`` will be used.
    If ``crawler`` is ``None``, only the ``from_settings`` constructor will be
    tried.

    ``*args`` and ``**kwargs`` are forwarded to the constructors.

    Raises ``ValueError`` if both ``settings`` and ``crawler`` are ``None``.

    .. versionchanged:: 2.2
       Raises ``TypeError`` if the resulting instance is ``None`` (e.g. if an
       extension has not been implemented correctly).
    """
    if settings is None:
        if crawler is None and spider is None:
            raise ValueError("Specify at least one of settings, crawler and spider.")

        settings = crawler and crawler.settings or spider and spider.settings
        spider = spider or crawler and crawler.spider

    if crawler and hasattr(objcls, 'from_crawler'):
        instance = await call_helper(objcls.from_crawler, crawler, *args, **kwargs)
        method_name = 'from_crawler'
    elif spider and hasattr(objcls, 'from_spider'):
        instance = await call_helper(objcls.from_spider, spider, *args, **kwargs)
        method_name = 'from_spider'
    elif hasattr(objcls, 'from_settings'):
        instance = await call_helper(objcls.from_settings, settings, *args, **kwargs)
        method_name = 'from_settings'
    else:
        instance = objcls(*args, **kwargs)
        method_name = '__new__'
    if instance is None:
        raise TypeError(f"{objcls.__qualname__}.{method_name} returned None")
    return instance


async def load_instance(clspath: str, *args, settings=None, spider=None, crawler=None, **kwargs):
    return await create_instance(
        load_object(clspath),
        settings,
        crawler,
        *args,
        spider=spider,
        **kwargs
    )



import argparse
import warnings
from shlex import split
from http.cookies import SimpleCookie
from urllib.parse import urlparse

from w3lib.http import basic_auth_header


class CurlParser(argparse.ArgumentParser):
    def error(self, message):
        error_msg = f'There was an error parsing the curl command: {message}'
        raise ValueError(error_msg)


curl_parser = CurlParser()
curl_parser.add_argument('url')
curl_parser.add_argument('-H', '--header', dest='headers', action='append')
curl_parser.add_argument('-X', '--request', dest='method')
curl_parser.add_argument('-d', '--data', '--data-raw', dest='data')
curl_parser.add_argument('-u', '--user', dest='auth')


safe_to_ignore_arguments = [
    ['--compressed'],
    # `--compressed` argument is not safe to ignore, but it's included here
    # because the `HttpCompressionMiddleware` is enabled by default
    ['-s', '--silent'],
    ['-v', '--verbose'],
    ['-#', '--progress-bar']
]

for argument in safe_to_ignore_arguments:
    curl_parser.add_argument(*argument, action='store_true')


def _parse_headers_and_cookies(parsed_args):
    headers = []
    cookies = {}
    for header in parsed_args.headers or ():
        name, val = header.split(':', 1)
        name = name.strip()
        val = val.strip()
        if name.title() == 'Cookie':
            for name, morsel in SimpleCookie(val).items():
                cookies[name] = morsel.value
        else:
            headers.append((name, val))

    if parsed_args.auth:
        user, password = parsed_args.auth.split(':', 1)
        headers.append(('Authorization', basic_auth_header(user, password)))

    return headers, cookies


def curl_to_request_kwargs(curl_command, ignore_unknown_options=True):
    """Convert a cURL command syntax to Request kwargs.

    :param str curl_command: string containing the curl command
    :param bool ignore_unknown_options: If true, only a warning is emitted when
                                        cURL options are unknown. Otherwise
                                        raises an error. (default: True)
    :return: dictionary of Request kwargs
    """

    curl_args = split(curl_command)

    if curl_args[0] != 'curl':
        raise ValueError('A curl command must start with "curl"')

    parsed_args, argv = curl_parser.parse_known_args(curl_args[1:])

    if argv:
        msg = f'Unrecognized options: {", ".join(argv)}'
        if ignore_unknown_options:
            warnings.warn(msg)
        else:
            raise ValueError(msg)

    url = parsed_args.url

    # curl automatically prepends 'http' if the scheme is missing, but Request
    # needs the scheme to work
    parsed_url = urlparse(url)
    if not parsed_url.scheme:
        url = 'http://' + url

    method = parsed_args.method or 'GET'

    result = {'method': method.upper(), 'url': url}

    headers, cookies = _parse_headers_and_cookies(parsed_args)

    if headers:
        result['headers'] = headers
    if cookies:
        result['cookies'] = cookies
    if parsed_args.data:
        result['body'] = parsed_args.data
        if not parsed_args.method:
            # if the "data" is specified but the "method" is not specified,
            # the default method is 'POST'
            result['method'] = 'POST'

    return result



import asyncio
import logging
import re
import ssl
from typing import Optional

import aiohttp

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings

logger = logging.getLogger(__name__)


class AioHttpDownloadHandler(BaseDownloadHandler):
    session: Optional[aiohttp.ClientSession] = None

    def __init__(self, settings: Settings):
        self.settings = settings
        self.aiohttp_client_session_args: dict = settings.getdict('AIOHTTP_CLIENT_SESSION_ARGS')
        self.verify_ssl: bool = settings.getbool("VERIFY_SSL", False)
        self.ssl_protocol = settings.get("SSL_PROTOCOL")  # ssl.PROTOCOL_TLSv1_2
        self.use_session: bool = settings.getbool("USE_SESSION", False)

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    def get_session(self, *args, **kwargs) -> aiohttp.ClientSession:
        if self.session is None:
            self.session = aiohttp.ClientSession(*args, **kwargs)
        return self.session

    async def download_request(self, request: Request, _) -> HtmlResponse:
        kwargs = {
            'verify_ssl': request.meta.get('verify_ssl', self.verify_ssl),
            'timeout': request.meta.get('download_timeout', 180),
            'cookies': dict(request.cookies),
            'data': request.body or None,
            'allow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect'),
            'max_redirects': self.settings.getint('REDIRECT_MAX_TIMES', 10),
        }

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        ssl_ciphers: str = request.meta.get('TLS_CIPHERS')
        ssl_protocol = request.meta.get('ssl_protocol', self.ssl_protocol)
        if ssl_ciphers or ssl_protocol:
            if ssl_protocol:
                context = ssl.SSLContext(protocol=ssl_protocol)
            else:
                context = ssl.create_default_context()

            ssl_ciphers and context.set_ciphers(ssl_ciphers)
            kwargs['ssl'] = context
            kwargs['verify_ssl'] = True

        proxy: str = request.meta.get("proxy")
        if proxy:
            kwargs["proxy"] = proxy
            logger.debug(f"使用代理{proxy}抓取: {request.url}")

        if self.use_session:
            # Not recommended to use session, The abnormal phenomena will occurs when using tunnel proxy
            session = self.get_session(**self.aiohttp_client_session_args)
            async with session.request(request.method, request.url, **kwargs) as response:
                content: bytes = await response.read()

        else:
            async with aiohttp.ClientSession(**self.aiohttp_client_session_args) as session:
                async with session.request(request.method, request.url, **kwargs) as response:
                    content: bytes = await response.read()

        r_cookies = response.cookies.output() or None
        if r_cookies:
            r_cookies = {
                cookie[0]: cookie[1] for cookie in re.findall(r'Set-Cookie: (.*?)=(.*?); Domain', r_cookies, re.S)
            }

        return HtmlResponse(
            str(response.url),
            status=response.status,
            headers=response.headers,
            body=content,
            cookies=r_cookies,
            encoding=response.charset
        )

    async def close(self):
        if self.session is not None:
            await self.session.close()

            # Wait 250 ms for the underlying SSL connections to close
            # https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
            await asyncio.sleep(0.250)



import os
import logging

from aioscrapy.utils.request import referer_str

SCRAPEDMSG = "Scraped from %(src)s" + os.linesep + "%(item)s"
DROPPEDMSG = "Dropped: %(exception)s" + os.linesep + "%(item)s"
CRAWLEDMSG = "Crawled (%(status)s) %(request)s%(request_flags)s (referer: %(referer)s)%(response_flags)s"
ITEMERRORMSG = "Error processing %(item)s"
SPIDERERRORMSG = "Spider error processing %(request)s (referer: %(referer)s)"
DOWNLOADERRORMSG_SHORT = "Error downloading %(request)s"
DOWNLOADERRORMSG_LONG = "Error downloading %(request)s: %(errmsg)s"


class LogFormatter:
    @staticmethod
    def crawled(request, response, spider):
        """Logs a message when the crawler finds a webpage."""
        request_flags = f' {str(request.flags)}' if request.flags else ''
        response_flags = f' {str(response.flags)}' if response.flags else ''
        return {
            'level': logging.DEBUG,
            'msg': CRAWLEDMSG,
            'args': {
                'status': response.status,
                'request': request,
                'request_flags': request_flags,
                'referer': referer_str(request),
                'response_flags': response_flags,
                'flags': response_flags
            }
        }

    @staticmethod
    def scraped(item, response, spider):
        """Logs a message when an item is scraped by a spider."""
        src = response
        return {
            'level': logging.DEBUG,
            'msg': SCRAPEDMSG,
            'args': {
                'src': src,
                'item': item,
            }
        }

    @staticmethod
    def dropped(item, exception, response, spider):
        """Logs a message when an item is dropped while it is passing through the item pipeline."""
        return {
            'level': logging.WARNING,
            'msg': DROPPEDMSG,
            'args': {
                'exception': exception,
                'item': item,
            }
        }

    @staticmethod
    def item_error(item, exception, response, spider):
        """Logs a message when an item causes an error while it is passing
        through the item pipeline.

        .. versionadded:: 2.0
        """
        return {
            'level': logging.ERROR,
            'msg': ITEMERRORMSG,
            'args': {
                'item': item,
            }
        }

    @staticmethod
    def spider_error(failure, request, response, spider):
        """Logs an error message from a spider.

        .. versionadded:: 2.0
        """
        return {
            'level': logging.ERROR,
            'msg': SPIDERERRORMSG,
            'args': {
                'request': request,
                'referer': referer_str(request),
            }
        }

    @staticmethod
    def download_error(failure, request, spider, errmsg=None):
        """Logs a download error message from a spider (typically coming from
        the engine).

        .. versionadded:: 2.0
        """
        args = {'request': request}
        if errmsg:
            msg = DOWNLOADERRORMSG_LONG
            args['errmsg'] = errmsg
        else:
            msg = DOWNLOADERRORMSG_SHORT
        return {
            'level': logging.ERROR,
            'msg': msg,
            'args': args,
        }

    @classmethod
    def from_crawler(cls, crawler):
        return cls()



import logging

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.http import PlaywrightResponse
from aioscrapy.settings import Settings
from .driverpool import WebDriverPool
from .webdriver import PlaywrightDriver

logger = logging.getLogger(__name__)


class PlaywrightHandler(BaseDownloadHandler):
    def __init__(self, settings: Settings):
        self.settings = settings
        playwright_client_args = settings.getdict('PLAYWRIGHT_CLIENT_ARGS')
        self.url_regexes = playwright_client_args.pop('url_regexes', [])
        pool_size = playwright_client_args.pop('pool_size', settings.getint("CONCURRENT_REQUESTS", 1))
        self._webdriver_pool = WebDriverPool(pool_size=pool_size, driver_cls=PlaywrightDriver, **playwright_client_args)

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, spider) -> PlaywrightResponse:
        cookies = dict(request.cookies)
        timeout = request.meta.get('download_timeout', 5) * 1000
        user_agent = (request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')).get("User-agent")
        proxy: str = request.meta.get("proxy")
        url = request.url

        driver: PlaywrightDriver = await self._webdriver_pool.get(
            user_agent=user_agent,
            proxy=proxy,
            timeout=timeout,
            on_event={
                name.replace('on_event', ''): getattr(spider, name) for name in dir(spider) if name.startswith('on_event')
            },
        )
        try:
            if cookies:
                driver.url = url
                await driver.set_cookies(cookies)
            await driver.page.goto(url, wait_until=request.meta.get('wait_until', "networkidle"))
            cache_response = {}
            for url_regex in self.url_regexes:
                async with driver.page.expect_response(url_regex, timeout=int(timeout/len(self.url_regexes))) as result:
                    res = await result.value
                    cache_response[url_regex] = PlaywrightResponse(
                        url=res.url,
                        request=request,
                        intercept_request=dict(
                            url=res.request.url,
                            headers=res.request.headers,
                            data=res.request.post_data,
                        ),
                        headers=res.headers,
                        body=await res.body(),
                        status=res.status,
                    )
            return PlaywrightResponse(
                url=driver.page.url,
                status=200,
                text=await driver.page.content(),
                cookies=await driver.get_cookies(),
                cache_response=cache_response,
                driver=driver,
                driver_pool=self._webdriver_pool
            )
        except Exception as e:
            await self._webdriver_pool.remove(driver)
            raise e

    async def close(self):
        await self._webdriver_pool.close()



from os.path import dirname, join
from setuptools import setup, find_packages

with open(join(dirname(__file__), 'aioscrapy/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()

install_requires = [
    "aiohttp",
    "ujson",
    "w3lib>=1.17.0",
    "parsel>=1.5.0",
    "PyDispatcher>=2.0.5",
    'zope.interface>=5.1.0',
    "redis>=4.3.1",
]
extras_require = {
    "all": [
        "aiomysql>=0.1.1", "httpx[http2]>=0.23.0", "aio-pika>=8.1.1",
        "cryptography", "motor>=3.1.1", "playwright>=1.31.1"
    ],
    "aiomysql": ["aiomysql>=0.1.1", "cryptography"],
    "httpx": ["httpx[http2]>=0.23.0"],
    "aio-pika": ["aio-pika>=8.1.1"],
    "mongo": ["motor>=3.1.1"],
    "playwright": ["playwright>=1.31.1"],
}

setup(
    name='aio-scrapy',
    version=version,
    url='https://github.com/conlin-huang/aio-scrapy.git',
    description='A high-level Web Crawling and Web Scraping framework based on Asyncio',
    long_description_content_type="text/markdown",
    long_description=open('README.md', encoding='utf-8').read(),
    author='conlin',
    author_email="995018884@qq.com",
    license="MIT",
    packages=find_packages(exclude=('example',)),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['aioscrapy = aioscrapy.cmdline:execute']
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.7',
    install_requires=install_requires,
    extras_require=extras_require,
    keywords=[
        'aio-scrapy',
        'scrapy',
        'aioscrapy',
        'scrapy redis',
        'asyncio',
        'spider',
    ]
)



import re
from typing import Optional

from playwright.async_api import Response

from aioscrapy.http.response.text import TextResponse


class PlaywrightResponse(TextResponse):
    def __init__(
            self,
            *args,
            text: str = '',
            cache_response: Optional[dict] = None,
            driver: Optional["PlaywrightDriver"] = None,
            driver_pool: Optional["WebDriverPool"] = None,
            intercept_request: Optional[dict] = None,
            **kwargs
    ):
        self.driver = driver
        self.driver_pool = driver_pool
        self._text = text
        self.cache_response = cache_response or {}
        self.intercept_request = intercept_request
        super().__init__(*args, **kwargs)

    async def release(self):
        self.driver_pool and self.driver and await self.driver_pool.release(self.driver)

    @property
    def text(self):
        return self._text or super().text

    @text.setter
    def text(self, text):
        self._text = text

    def get_response(self, url_regex):
        return self.cache_response.get(url_regex)



# -*- coding: utf-8 -*-

from asyncio import Lock
from asyncio.queues import Queue

from aioscrapy.utils.tools import singleton


@singleton
class WebDriverPool:
    def __init__(
            self, pool_size=5, driver_cls=None, **kwargs
    ):
        self.pool_size = pool_size
        self.driver_cls = driver_cls
        self.kwargs = kwargs

        self.queue = Queue(maxsize=pool_size)
        self.lock = Lock()
        self.driver_count = 0

    @property
    def is_full(self):
        return self.driver_count >= self.pool_size

    async def create_driver(self, **args):
        kwargs = self.kwargs.copy()
        kwargs.update(args)
        driver = self.driver_cls(**kwargs)
        await driver.setup()
        return driver

    async def get(self, **kwargs):
        async with self.lock:
            if not self.is_full:
                driver = await self.create_driver(**kwargs)
                self.driver_count += 1
            else:
                driver = await self.queue.get()
        return driver

    async def release(self, driver):
        await self.queue.put(driver)

    async def remove(self, driver):
        await driver.quit()
        self.driver_count -= 1

    async def close(self):
        while not self.queue.empty():
            driver = await self.queue.get()
            await driver.quit()
            self.driver_count -= 1
